Compare commits
2 Commits
9ca33cf4d5
...
09e5261ba6
| Author | SHA1 | Date | |
|---|---|---|---|
|
|
09e5261ba6 | ||
|
|
ab42fc1fd7 |
156
ex_engine/xllm_kernels/build_test_hgemm.sh
Executable file
156
ex_engine/xllm_kernels/build_test_hgemm.sh
Executable file
@@ -0,0 +1,156 @@
|
||||
#!/bin/bash
|
||||
# build_test_hgemm.sh — Compile and test hgemm_blocktiling on BI-V100
|
||||
#
|
||||
# Usage: bash ex_engine/xllm_kernels/build_test_hgemm.sh
|
||||
set -eo pipefail
|
||||
|
||||
SCRIPT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
|
||||
CUDA_DIR="${SCRIPT_DIR}/cuda"
|
||||
|
||||
echo "=== 1. Compile hgemm_blocktiling ==="
|
||||
python3 -c "
|
||||
import torch.utils.cpp_extension as ext
|
||||
import os, shutil, glob
|
||||
|
||||
name = 'hgemm_blocktiling'
|
||||
build_dir = '${SCRIPT_DIR}/build/tmp_' + name
|
||||
os.makedirs(build_dir, exist_ok=True)
|
||||
|
||||
try:
|
||||
mod = ext.load(
|
||||
name=name,
|
||||
sources=[
|
||||
'${CUDA_DIR}/hgemm_blocktiling.cu',
|
||||
'${CUDA_DIR}/bindings/hgemm_bind.cpp',
|
||||
],
|
||||
extra_include_paths=['${CUDA_DIR}/headers'],
|
||||
extra_cflags=['-O2', '-std=c++17'],
|
||||
extra_cuda_cflags=['-O2'],
|
||||
build_directory=build_dir,
|
||||
verbose=True,
|
||||
)
|
||||
built = glob.glob(build_dir + '/' + name + '*.so')
|
||||
if built:
|
||||
dst = '${SCRIPT_DIR}/build/' + name + '.so'
|
||||
shutil.copy2(built[0], dst)
|
||||
print(f'[build] SUCCESS: {dst} ({os.path.getsize(dst)} bytes)')
|
||||
else:
|
||||
print('[build] WARNING: .so not found')
|
||||
except Exception as e:
|
||||
print(f'[build] FAILED: {e}')
|
||||
import traceback
|
||||
traceback.print_exc()
|
||||
"
|
||||
|
||||
echo ""
|
||||
echo "=== 2. Functional test ==="
|
||||
python3 << 'PYTEST'
|
||||
import torch
|
||||
import sys, os, glob
|
||||
|
||||
# Find and load the .so
|
||||
build_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)) if '__file__' in dir() else '.',
|
||||
'ex_engine/xllm_kernels/build')
|
||||
sys.path.insert(0, build_dir)
|
||||
|
||||
try:
|
||||
import hgemm_blocktiling as hg
|
||||
print("Module loaded successfully")
|
||||
except ImportError:
|
||||
# Try loading from tmp build dir
|
||||
import importlib.util
|
||||
so_files = glob.glob('ex_engine/xllm_kernels/build/tmp_hgemm_blocktiling/hgemm_blocktiling*.so')
|
||||
if not so_files:
|
||||
print("SKIP: .so not found (need GPU machine)")
|
||||
sys.exit(0)
|
||||
spec = importlib.util.spec_from_file_location("hgemm_blocktiling", so_files[0])
|
||||
hg = importlib.util.module_from_spec(spec)
|
||||
spec.loader.exec_module(hg)
|
||||
print(f"Module loaded from {so_files[0]}")
|
||||
|
||||
# Test 1: Small GEMM correctness
|
||||
print("\n--- Test 1: Small GEMM (64x64 @ 64x64) ---")
|
||||
M, N, K = 64, 64, 64
|
||||
A = torch.randn(M, K, dtype=torch.float16, device='cuda')
|
||||
B = torch.randn(K, N, dtype=torch.float16, device='cuda')
|
||||
|
||||
C_ref = torch.matmul(A.float(), B.float()).half()
|
||||
C_our = hg.hgemm(A, B)
|
||||
|
||||
diff = (C_ref.float() - C_our.float()).abs().max().item()
|
||||
print(f" Max abs diff: {diff:.6f}")
|
||||
assert diff < 1.0, f"FAILED: diff={diff} too large"
|
||||
print(f" PASS (diff < 1.0)")
|
||||
|
||||
# Test 2: Larger GEMM (typical MoE dimensions)
|
||||
print("\n--- Test 2: MoE-sized GEMM (256x4096 @ 4096x11008) ---")
|
||||
M, N, K = 256, 11008, 4096
|
||||
A = torch.randn(M, K, dtype=torch.float16, device='cuda') * 0.01
|
||||
B = torch.randn(K, N, dtype=torch.float16, device='cuda') * 0.01
|
||||
|
||||
C_ref = torch.matmul(A.float(), B.float()).half()
|
||||
C_our = hg.hgemm(A, B)
|
||||
|
||||
diff = (C_ref.float() - C_our.float()).abs().max().item()
|
||||
rel_diff = diff / (C_ref.float().abs().max().item() + 1e-8)
|
||||
print(f" Max abs diff: {diff:.6f}, rel: {rel_diff:.6f}")
|
||||
assert rel_diff < 0.05, f"FAILED: rel_diff={rel_diff} too large"
|
||||
print(f" PASS")
|
||||
|
||||
# Test 3: MoE expert GEMM with variable counts
|
||||
print("\n--- Test 3: MoE expert GEMM (8 experts, variable tokens) ---")
|
||||
num_experts = 8
|
||||
K_dim = 128
|
||||
N_dim = 256
|
||||
expert_counts = torch.tensor([32, 16, 0, 48, 8, 24, 4, 12], dtype=torch.int32)
|
||||
total_tokens = expert_counts.sum().item()
|
||||
|
||||
input_tensor = torch.randn(total_tokens, K_dim, dtype=torch.float16, device='cuda') * 0.1
|
||||
weights = torch.randn(num_experts, N_dim, K_dim, dtype=torch.float16, device='cuda') * 0.1
|
||||
|
||||
output = hg.moe_expert_gemm(input_tensor, weights, expert_counts.cuda())
|
||||
|
||||
# Verify against torch reference
|
||||
offset = 0
|
||||
for e in range(num_experts):
|
||||
cnt = expert_counts[e].item()
|
||||
if cnt == 0:
|
||||
continue
|
||||
inp_e = input_tensor[offset:offset+cnt]
|
||||
w_e = weights[e] # (N, K)
|
||||
ref_e = torch.matmul(inp_e.float(), w_e.float().t()).half()
|
||||
out_e = output[offset:offset+cnt]
|
||||
diff_e = (ref_e.float() - out_e.float()).abs().max().item()
|
||||
print(f" Expert {e} (tokens={cnt}): max_diff={diff_e:.6f}")
|
||||
offset += cnt
|
||||
print(f" PASS")
|
||||
|
||||
# Test 4: Performance benchmark
|
||||
print("\n--- Test 4: Performance (256x4096 @ 4096x11008, 100 iters) ---")
|
||||
M, N, K = 256, 11008, 4096
|
||||
A = torch.randn(M, K, dtype=torch.float16, device='cuda')
|
||||
B = torch.randn(K, N, dtype=torch.float16, device='cuda')
|
||||
|
||||
# Warmup
|
||||
for _ in range(10):
|
||||
hg.hgemm(A, B)
|
||||
torch.cuda.synchronize()
|
||||
|
||||
import time
|
||||
start = time.time()
|
||||
for _ in range(100):
|
||||
hg.hgemm(A, B)
|
||||
torch.cuda.synchronize()
|
||||
elapsed = time.time() - start
|
||||
print(f" Custom kernel: {elapsed*10:.2f} ms/iter")
|
||||
|
||||
start = time.time()
|
||||
for _ in range(100):
|
||||
torch.matmul(A, B)
|
||||
torch.cuda.synchronize()
|
||||
elapsed2 = time.time() - start
|
||||
print(f" torch.matmul: {elapsed2*10:.2f} ms/iter")
|
||||
print(f" Ratio: {elapsed/elapsed2:.2f}x")
|
||||
|
||||
print("\n=== ALL TESTS PASSED ===")
|
||||
PYTEST
|
||||
132
ex_engine/xllm_kernels/cuda/bindings/hgemm_bind.cpp
Normal file
132
ex_engine/xllm_kernels/cuda/bindings/hgemm_bind.cpp
Normal file
@@ -0,0 +1,132 @@
|
||||
// hgemm_bind.cpp — pybind11 bindings for hgemm_blocktiling.cu
|
||||
//
|
||||
// Exports:
|
||||
// hgemm(A, B, M, N, K) → C
|
||||
// moe_expert_gemm(input, weights, expert_counts) → output
|
||||
|
||||
#include <torch/extension.h>
|
||||
#include <cuda_fp16.h>
|
||||
#include <vector>
|
||||
|
||||
// Forward declarations from hgemm_blocktiling.cu
|
||||
void launch_hgemm_blocktiling(
|
||||
int M, int N, int K,
|
||||
const __half* alpha, const __half* A, int lda,
|
||||
const __half* B, int ldb,
|
||||
const __half* beta, __half* C, int ldc,
|
||||
cudaStream_t stream);
|
||||
|
||||
void launch_moe_expert_hgemm(
|
||||
int num_experts,
|
||||
const int* expert_counts,
|
||||
const int* expert_offsets,
|
||||
int N, int K,
|
||||
const __half* input,
|
||||
const __half* weights,
|
||||
__half* output,
|
||||
cudaStream_t stream);
|
||||
|
||||
|
||||
// ============================================================================
|
||||
// Python-facing wrappers
|
||||
// ============================================================================
|
||||
|
||||
// Simple GEMM: C = A @ B
|
||||
// A: (M, K) fp16, B: (K, N) fp16 → C: (M, N) fp16
|
||||
torch::Tensor hgemm(torch::Tensor A, torch::Tensor B) {
|
||||
TORCH_CHECK(A.is_cuda() && B.is_cuda(), "Inputs must be CUDA tensors");
|
||||
TORCH_CHECK(A.scalar_type() == torch::kHalf, "A must be fp16");
|
||||
TORCH_CHECK(B.scalar_type() == torch::kHalf, "B must be fp16");
|
||||
TORCH_CHECK(A.dim() == 2 && B.dim() == 2, "A and B must be 2D");
|
||||
TORCH_CHECK(A.size(1) == B.size(0), "Inner dimensions must match");
|
||||
|
||||
int M = A.size(0);
|
||||
int K = A.size(1);
|
||||
int N = B.size(1);
|
||||
|
||||
auto C = torch::zeros({M, N}, A.options());
|
||||
|
||||
__half alpha = __float2half(1.0f);
|
||||
__half beta = __float2half(0.0f);
|
||||
|
||||
cudaStream_t stream = at::cuda::getCurrentCUDAStream();
|
||||
|
||||
launch_hgemm_blocktiling(
|
||||
M, N, K, &alpha,
|
||||
reinterpret_cast<const __half*>(A.data_ptr<at::Half>()),
|
||||
A.size(1),
|
||||
reinterpret_cast<const __half*>(B.data_ptr<at::Half>()),
|
||||
B.size(1),
|
||||
&beta,
|
||||
reinterpret_cast<__half*>(C.data_ptr<at::Half>()),
|
||||
C.size(1),
|
||||
stream);
|
||||
|
||||
return C;
|
||||
}
|
||||
|
||||
|
||||
// MoE expert GEMM: for each expert e, compute
|
||||
// output[offset_e : offset_e + count_e] = input[offset_e : offset_e + count_e] @ weights[e].T
|
||||
//
|
||||
// input: (total_tokens, K) fp16
|
||||
// weights: (num_experts, N, K) fp16 — weight layout matches vllm w13/w2 convention
|
||||
// expert_counts: (num_experts,) int32 — number of tokens per expert
|
||||
//
|
||||
// Returns: output (total_tokens, N) fp16
|
||||
torch::Tensor moe_expert_gemm(
|
||||
torch::Tensor input,
|
||||
torch::Tensor weights,
|
||||
torch::Tensor expert_counts
|
||||
) {
|
||||
TORCH_CHECK(input.is_cuda() && weights.is_cuda(), "Inputs must be CUDA");
|
||||
TORCH_CHECK(input.scalar_type() == torch::kHalf, "input must be fp16");
|
||||
TORCH_CHECK(weights.scalar_type() == torch::kHalf, "weights must be fp16");
|
||||
TORCH_CHECK(expert_counts.scalar_type() == torch::kInt32 ||
|
||||
expert_counts.scalar_type() == torch::kInt64,
|
||||
"expert_counts must be int32 or int64");
|
||||
|
||||
int total_tokens = input.size(0);
|
||||
int K = input.size(1);
|
||||
int num_experts = weights.size(0);
|
||||
int N = weights.size(1); // output dim
|
||||
|
||||
TORCH_CHECK(weights.size(2) == K, "weights K dim must match input");
|
||||
|
||||
auto output = torch::zeros({total_tokens, N}, input.options());
|
||||
|
||||
// Convert expert_counts to host int array
|
||||
auto counts_cpu = expert_counts.to(torch::kCPU).to(torch::kInt32).contiguous();
|
||||
std::vector<int> counts(num_experts);
|
||||
std::vector<int> offsets(num_experts);
|
||||
int cumsum = 0;
|
||||
for (int i = 0; i < num_experts; i++) {
|
||||
counts[i] = counts_cpu.data_ptr<int32_t>()[i];
|
||||
offsets[i] = cumsum;
|
||||
cumsum += counts[i];
|
||||
}
|
||||
|
||||
cudaStream_t stream = at::cuda::getCurrentCUDAStream();
|
||||
|
||||
launch_moe_expert_hgemm(
|
||||
num_experts,
|
||||
counts.data(),
|
||||
offsets.data(),
|
||||
N, K,
|
||||
reinterpret_cast<const __half*>(input.data_ptr<at::Half>()),
|
||||
reinterpret_cast<const __half*>(weights.data_ptr<at::Half>()),
|
||||
reinterpret_cast<__half*>(output.data_ptr<at::Half>()),
|
||||
stream);
|
||||
|
||||
return output;
|
||||
}
|
||||
|
||||
|
||||
PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
|
||||
m.def("hgemm", &hgemm,
|
||||
"FP16 GEMM: C = A @ B (adapted from siboehm kernel 6 for BI-V100)",
|
||||
py::arg("A"), py::arg("B"));
|
||||
m.def("moe_expert_gemm", &moe_expert_gemm,
|
||||
"MoE expert GEMM: per-expert matmul with variable token counts",
|
||||
py::arg("input"), py::arg("weights"), py::arg("expert_counts"));
|
||||
}
|
||||
167
ex_engine/xllm_kernels/cuda/hgemm_blocktiling.cu
Normal file
167
ex_engine/xllm_kernels/cuda/hgemm_blocktiling.cu
Normal file
@@ -0,0 +1,167 @@
|
||||
// hgemm_blocktiling.cu — FP16 GEMM for BI-V100
|
||||
//
|
||||
// 1:1 from siboehm/SGEMM_CUDA kernel 6 (sgemmVectorize).
|
||||
// Changes: float→__half, float4→load 4 halfs, FP32 accumulator.
|
||||
// No WARPSIZE usage. No cooperative_groups. CUDA 10.2 safe.
|
||||
|
||||
#include <cuda_fp16.h>
|
||||
#include <cuda_runtime.h>
|
||||
|
||||
#define CEIL_DIV(M, N) (((M) + (N)-1) / (N))
|
||||
|
||||
template <const int BM, const int BN, const int BK, const int TM, const int TN>
|
||||
__global__ void hgemmVectorize(int M, int N, int K, float alpha,
|
||||
const __half *A, const __half *B,
|
||||
float beta, __half *C) {
|
||||
const uint cRow = blockIdx.y;
|
||||
const uint cCol = blockIdx.x;
|
||||
|
||||
// BN/TN are the number of threads to span a column
|
||||
const int threadCol = threadIdx.x % (BN / TN);
|
||||
const int threadRow = threadIdx.x / (BN / TN);
|
||||
|
||||
// allocate space for the current blocktile in smem
|
||||
// A stored transposed: As[BK][BM], B normal: Bs[BK][BN]
|
||||
__shared__ __half As[BM * BK];
|
||||
__shared__ __half Bs[BK * BN];
|
||||
|
||||
// Move blocktile to beginning of A's row and B's column
|
||||
A += cRow * BM * K;
|
||||
B += cCol * BN;
|
||||
C += cRow * BM * N + cCol * BN;
|
||||
|
||||
// calculating the indices that this thread will load into SMEM
|
||||
// FP16: load 4 halfs (8 bytes) per step. 4 halfs per thread.
|
||||
// siboehm: float4 = 4 floats = 128bit. We do 4 halfs = 64bit.
|
||||
const uint innerRowA = threadIdx.x / (BK / 4);
|
||||
const uint innerColA = threadIdx.x % (BK / 4);
|
||||
const uint innerRowB = threadIdx.x / (BN / 4);
|
||||
const uint innerColB = threadIdx.x % (BN / 4);
|
||||
|
||||
// allocate thread-local cache for results in registerfile
|
||||
// FP32 accumulation to avoid FP16 precision loss
|
||||
float threadResults[TM * TN] = {0.0f};
|
||||
__half regM[TM];
|
||||
__half regN[TN];
|
||||
|
||||
// outer-most loop over block tiles
|
||||
for (uint bkIdx = 0; bkIdx < K; bkIdx += BK) {
|
||||
// populate the SMEM caches
|
||||
// transpose A while loading it (same as siboehm)
|
||||
// Load 4 halfs from A
|
||||
__half a0 = A[innerRowA * K + innerColA * 4 + 0];
|
||||
__half a1 = A[innerRowA * K + innerColA * 4 + 1];
|
||||
__half a2 = A[innerRowA * K + innerColA * 4 + 2];
|
||||
__half a3 = A[innerRowA * K + innerColA * 4 + 3];
|
||||
As[(innerColA * 4 + 0) * BM + innerRowA] = a0;
|
||||
As[(innerColA * 4 + 1) * BM + innerRowA] = a1;
|
||||
As[(innerColA * 4 + 2) * BM + innerRowA] = a2;
|
||||
As[(innerColA * 4 + 3) * BM + innerRowA] = a3;
|
||||
|
||||
// Load 4 halfs from B (no transpose)
|
||||
Bs[innerRowB * BN + innerColB * 4 + 0] = B[innerRowB * N + innerColB * 4 + 0];
|
||||
Bs[innerRowB * BN + innerColB * 4 + 1] = B[innerRowB * N + innerColB * 4 + 1];
|
||||
Bs[innerRowB * BN + innerColB * 4 + 2] = B[innerRowB * N + innerColB * 4 + 2];
|
||||
Bs[innerRowB * BN + innerColB * 4 + 3] = B[innerRowB * N + innerColB * 4 + 3];
|
||||
__syncthreads();
|
||||
|
||||
// advance blocktile
|
||||
A += BK; // move BK columns to right
|
||||
B += BK * N; // move BK rows down
|
||||
|
||||
// calculate per-thread results
|
||||
for (uint dotIdx = 0; dotIdx < BK; ++dotIdx) {
|
||||
// block into registers
|
||||
for (uint i = 0; i < TM; ++i) {
|
||||
regM[i] = As[dotIdx * BM + threadRow * TM + i];
|
||||
}
|
||||
for (uint i = 0; i < TN; ++i) {
|
||||
regN[i] = Bs[dotIdx * BN + threadCol * TN + i];
|
||||
}
|
||||
// FP32 accumulation
|
||||
for (uint resIdxM = 0; resIdxM < TM; ++resIdxM) {
|
||||
float aVal = __half2float(regM[resIdxM]);
|
||||
for (uint resIdxN = 0; resIdxN < TN; ++resIdxN) {
|
||||
threadResults[resIdxM * TN + resIdxN] +=
|
||||
aVal * __half2float(regN[resIdxN]);
|
||||
}
|
||||
}
|
||||
}
|
||||
__syncthreads();
|
||||
}
|
||||
|
||||
// write out the results
|
||||
for (uint resIdxM = 0; resIdxM < TM; resIdxM += 1) {
|
||||
for (uint resIdxN = 0; resIdxN < TN; resIdxN += 1) {
|
||||
uint row = cRow * BM + threadRow * TM + resIdxM;
|
||||
uint col = cCol * BN + threadCol * TN + resIdxN;
|
||||
if (row < M && col < N) {
|
||||
float c_old = __half2float(C[(threadRow * TM + resIdxM) * N +
|
||||
threadCol * TN + resIdxN]);
|
||||
C[(threadRow * TM + resIdxM) * N + threadCol * TN + resIdxN] =
|
||||
__float2half(alpha * threadResults[resIdxM * TN + resIdxN] +
|
||||
beta * c_old);
|
||||
}
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
|
||||
// ============================================================================
|
||||
// Launch wrapper — matches siboehm runSgemmVectorize
|
||||
// ============================================================================
|
||||
void launch_hgemm_blocktiling(
|
||||
int M, int N, int K,
|
||||
const __half* alpha_ptr,
|
||||
const __half* A, int lda,
|
||||
const __half* B, int ldb,
|
||||
const __half* beta_ptr,
|
||||
__half* C, int ldc,
|
||||
cudaStream_t stream)
|
||||
{
|
||||
constexpr int BM = 128;
|
||||
constexpr int BN = 128;
|
||||
constexpr int BK = 8;
|
||||
constexpr int TM = 8;
|
||||
constexpr int TN = 8;
|
||||
// 256 threads — same as siboehm
|
||||
constexpr int NUM_THREADS = (BM * BN) / (TM * TN);
|
||||
|
||||
dim3 grid(CEIL_DIV(N, BN), CEIL_DIV(M, BM));
|
||||
dim3 block(NUM_THREADS);
|
||||
|
||||
float alpha = 1.0f, beta = 0.0f;
|
||||
if (alpha_ptr) alpha = __half2float(*alpha_ptr);
|
||||
if (beta_ptr) beta = __half2float(*beta_ptr);
|
||||
|
||||
hgemmVectorize<BM, BN, BK, TM, TN>
|
||||
<<<grid, block, 0, stream>>>(M, N, K, alpha, A, B, beta, C);
|
||||
}
|
||||
|
||||
|
||||
// ============================================================================
|
||||
// MoE expert GEMM — C++ loop over experts (replaces Python for-loop)
|
||||
// ============================================================================
|
||||
void launch_moe_expert_hgemm(
|
||||
int num_experts,
|
||||
const int* expert_counts, // host, [num_experts]
|
||||
const int* expert_offsets, // host, [num_experts]
|
||||
int N, int K,
|
||||
const __half* input, // (total_tokens, K)
|
||||
const __half* weights, // (num_experts, N, K)
|
||||
__half* output, // (total_tokens, N)
|
||||
cudaStream_t stream)
|
||||
{
|
||||
for (int e = 0; e < num_experts; e++) {
|
||||
int M_e = expert_counts[e];
|
||||
if (M_e == 0) continue;
|
||||
|
||||
int off = expert_offsets[e];
|
||||
const __half* A = input + off * K;
|
||||
const __half* B = weights + (long long)e * N * K;
|
||||
__half* C_e = output + off * N;
|
||||
|
||||
launch_hgemm_blocktiling(M_e, N, K,
|
||||
nullptr, A, K, B, N, nullptr, C_e, N, stream);
|
||||
}
|
||||
}
|
||||
Reference in New Issue
Block a user